"""
Has a state class storing all information on GUI state, and has
GTK classes representing data from a SpectrumContainer and the GUI state.
"""
# pylint: disable=wrong-import-position
# pylint: disable=logging-format-interpolation

import logging
import string

from bidict import OrderedBidict

from gxps.config import CONFIG
from gxps.utility import Observable #, EventQueue


LOG = logging.getLogger(__name__)


class State(Observable):
    """Holds information on the GUI state.
    """
    # pylint: disable=too-many-instance-attributes
    _signals = (
        "changed-plotting-options",
        "changed-project",
        "changed-tv",
        "changed-rsf",
        "changed-active",
        "changed-editing-spectra"
    )
    peak_name_list = [
        *list(string.ascii_uppercase),
        *[i+b for i in string.ascii_uppercase for b in string.ascii_uppercase]
    ]
    titles = {
        "spectrum_view": OrderedBidict({
            "name": "Name",
            "notes": "Notes"
        }),
        "peak_view": OrderedBidict({
            "label": "Label",
            "name": "     ",
            "shape": "Shape",
            "position": "Position",
            "area": "Area*",
            "fwhm": "FWHM*",
            "alpha": "Par1",
            "beta": "Par2",
            "gamma": "Par3"
        }),
        "static_specinfo": OrderedBidict({
            "filename": "Filename"
        }),
        "editing_dialog": OrderedBidict({
            "name": "Name",
            "notes": "Notes",
            "pass_energy": "Pass Energy",
            "integration_time": "Time per Data Point",
            "sweeps": "Sweeps"
        }),
        "norm_types": OrderedBidict({
            "none": "none",
            "manual": "manual",
            "high_energy": "high energy background",
            "low_energy": "low energy background",
            "highest": "highest peak"
        }),
        "norm_type_ids": OrderedBidict({
            "none": "0",
            "highest": "1",
            "high_energy": "2",
            "low_energy": "3",
            "manual": "4"
        }),
        "background_types": OrderedBidict({
            "none": "none",
            "shirley": "Shirley",
            "linear": "Linear"
        }),
        "background_type_ids": OrderedBidict({
            "none": "0",
            "shirley": "1",
            "linear": "2"
        }),
        "photon_source_ids": OrderedBidict({
            "Al": "0",
            "Mg": "1"
        })
    }

    def __init__(self, app, spectra):
        self._app = app
        self._spectra = spectra
        # Most of the states are properties: the direct values are non-public
        # Active data objects: Where all operations and visualization apply
        self._active_spectra = []
        self._editing_spectra = []
        self._active_peaks = []
        # Project file-related
        self._current_project = ""
        self._project_isaltered = False
        # Spectrum Treeview-related
        self._spectra_tv_columns = ["name", "notes"]
        self._spectra_tv_filter = ["notes", None]
        # Peak Trewwvie-related
        self._peak_tv_columns = [
            "name", "label", "position", "fwhm", "area",
            "alpha"
        ]
        # Plotting-related
        self._rsf_elements = [""]
        self._photon_source = "Al"
        # Option dicts
        self._plotting_options = {
            "show-spectra": True,
            "show-background-definition": True,
            "show-background": True,
            "subtract-background": False,
            "show-peaks": True,
            "show-rsfs": True
        }
        # Keep the active spectra up to date
        self._app.bus.subscribe(
            self.update_active, "changed-spectra", priority=0)
        self._app.bus.subscribe(
            self.update_active, "changed-fit", priority=0)
        signals = (
            "changed-spectra",
            "changed-spectrum",
            "changed-spectrum-meta",
            "changed-fit",
            "changed-peak",
            "changed-peak-meta"
        )
        for signal in signals:
            self._app.bus.subscribe(self.alter_project, signal, priority=10)

        super().__init__()

    @property
    def active_spectra(self):
        """Currently active spectra.
        """
        return self._active_spectra.copy()

    @active_spectra.setter
    def active_spectra(self, spectra):
        """Activates given spectra.
        """
        if spectra == self._active_spectra:
            return
        for spectrum in spectra.copy():
            if spectrum not in self._spectra.spectra:
                raise ValueError("Invalid spectrum activated.")
        self._active_spectra.clear()
        self._active_spectra.extend(spectra)
        self.emit("changed-active", attr="spectra")
        # update active peaks:
        peaks = self._active_peaks.copy()
        for peak in self._active_peaks:
            for spectrum in self._active_spectra:
                if peak in spectrum.peaks:
                    break
            else:
                peaks.remove(peak)
        self.active_peaks = peaks

    @property
    def selected_spectra(self):
        """Returns spectra currently selected in the spectrum_view."""
        selection = self._app.builder.get_object("spectrum_selection")
        model, pathlist = selection.get_selected_rows()
        iters = [model.get_iter(path) for path in pathlist]
        spectra = [model.get(iter_, 0)[0] for iter_ in iters]
        return spectra

    @property
    def editing_spectra(self):
        """Returns list of spectra which are currently in editing."""
        return self._editing_spectra.copy()

    @editing_spectra.setter
    def editing_spectra(self, spectra):
        """Set new spectra to edit. Emits 'changed-editing-spectra'."""
        if spectra == self._editing_spectra:
            return
        for spectrum in spectra.copy():
            if spectrum not in self._spectra.spectra:
                raise ValueError("Invalid spectrum activated.")
        self._editing_spectra.clear()
        self._editing_spectra.extend(spectra)
        self.emit("changed-editing-spectra")

    @property
    def visible_peaks(self):
        """Currently visible peaks: the ones that belong to active spectra.
        """
        v_peaks = []
        for spectrum in self.active_spectra:
            v_peaks.extend(spectrum.peaks)
        return v_peaks

    @property
    def active_peaks(self):
        """Currently active peaks.
        """
        # for peak in self._active_peaks:
        #     for spectrum in self.active_spectra:
        #         if peak in spectrum.peaks:
        #             break
        #     else:
        #         self._active_peaks.remove(peak)
        return self._active_peaks.copy()

    @active_peaks.setter
    def active_peaks(self, peaks):
        """Activates given peaks.
        """
        if peaks == self._active_peaks:
            return
        for peak in peaks.copy():
            for spectrum in self.active_spectra:
                if peak in spectrum.peaks:
                    break
            else:
                raise ValueError("Invalid peak activated.")
        self._active_peaks.clear()
        self._active_peaks.extend(peaks)
        self.emit("changed-active", attr="peaks")

    @property
    def peak_names(self):
        """Already taken peak names."""
        peak_names = []
        for spectrum in self._spectra.spectra:
            for peak in spectrum.peaks:
                peak_names.append(peak.name)
        return peak_names

    @property
    def next_peak_name(self):
        """Returns the next peak name that is free."""
        peak_names = self.peak_names
        for name in self.peak_name_list:
            if name not in peak_names:
                return name
        return "N/A"

    @property
    def selected_peaks(self):
        """Currently selected peaks. Will be set to active peaks
        immediately by the ViewController.
        """
        selection = self._app.builder.get_object("peak_selection")
        model, pathlist = selection.get_selected_rows()
        iters = [model.get_iter(path) for path in pathlist]
        peaks = [model.get(iter_, 0)[0] for iter_ in iters]
        return peaks

    def update_active(self, *_args):
        """Clean out active spectra/peaks that do not exist anymore."""
        spectra = self._active_spectra.copy()
        for spectrum in self._active_spectra:
            if spectrum not in self._spectra.spectra:
                spectra.remove(spectrum)
        self.active_spectra = spectra
        peaks = self._active_peaks.copy()
        for peak in self._active_peaks:
            for spectrum in self._active_spectra:
                if peak in spectrum.peaks:
                    break
            else:
                peaks.remove(peak)
        self.active_peaks = peaks

    @property
    def current_project(self):
        """Filename of the currently opened project."""
        if self._current_project == "Untitled":
            return ""
        return self._current_project

    @current_project.setter
    def current_project(self, value):
        """Set the current project file name, also in the config."""
        if value == "Untitled" or not value:
            self._current_project = "Untitled"
            CONFIG["IO"]["current-project"] = ""
            self._project_isaltered = False
            self.emit("changed-project", attr="filename")
        elif value != self._current_project:
            self._current_project = value
            CONFIG["IO"]["current-project"] = value
            self._project_isaltered = False
            self.emit("changed-project", attr="filename")
        else:
            self._project_isaltered = False
            self.emit("changed-project", attr="filename")

    @property
    def project_isaltered(self):
        """Boolean if the project was altered since last save.
        """
        return self._project_isaltered

    @project_isaltered.setter
    def project_isaltered(self, isaltered):
        """Project was altered or saved."""
        if isaltered != self._project_isaltered:
            self._project_isaltered = isaltered
            self.emit("changed-project", attr="isaltered")

    def alter_project(self, _event):
        """Helper for setting project_isaltered."""
        self.project_isaltered = True

    @property
    def spectra_tv_columns(self):
        """Treeview columns for spectra view.
        """
        return self._spectra_tv_columns.copy()

    @property
    def spectra_tv_filter(self):
        """Treeview filter for spectra view.
        """
        return self._spectra_tv_filter.copy()

    @spectra_tv_filter.setter
    def spectra_tv_filter(self, value):
        """Set Spectrum Treeview Filter.
        """
        try:
            meta_attr_title, search_term = value
            meta_attr = self.titles["spectrum_view"].inverse[meta_attr_title]
            self._spectra_tv_filter[0] = meta_attr
            self._spectra_tv_filter[1] = search_term
            self.emit("changed-tv", attr="filter")
        except TypeError:
            LOG.error("Spectrum treeview filter '{}' invalid".format(value))

    @property
    def peak_tv_columns(self):
        """Treeview columns for peaks view.
        """
        return self._peak_tv_columns.copy()

    @property
    def rsf_elements(self):
        """Elements for which RSF values are plotted."""
        return self._rsf_elements.copy()

    @rsf_elements.setter
    def rsf_elements(self, elements):
        """Set Elements for which RSF values should be plotted."""
        self._rsf_elements.clear()
        self._rsf_elements.extend(elements)
        self.emit("changed-rsf", attr="elements")

    @property
    def photon_source(self):
        """Currently selected photon source."""
        return self._photon_source

    @property
    def photon_source_id(self):
        """ID of the active photon source."""
        return self.titles["photon_source_ids"][self._photon_source]

    @photon_source.setter
    def photon_source(self, source):
        """Set photon source."""
        self._photon_source = source
        self.emit("changed-rsf", attr="source")

    @property
    def plotting_options(self):
        """Option dictionary for how to plot the spectra and peaks."""
        return self._plotting_options.copy()

    def set_plotting_option(self, **options):
        """Sets plotting options via keyword arguments."""
        self._plotting_options.update(options)
        self.emit("changed-plotting-options")



"""Tests the gui module."""
# pylint: disable=missing-docstring

# import pytest

from gxps import gui


def test_gui_state():
    state = gui.GUIState()
    state.active_peak = "p1"
    state2 = gui.GUIState()
    assert state2.active_peak == "p1"
    gui.GUIState.cleanup()



"""Loads the configuration ini and sets app name and version."""
# pylint: disable=invalid-name
# pylint: disable=wrong-import-order
# pylint: disable=logging-format-interpolation

__appname__ = "GXPS"
__authors__ = ["Simon Fischer <sf@simon-fischer.info>"]
__website__ = "https://github.com/schachmett/gxps"

# maybe use release_string() instead?
__version__ = "unknown version"
__release__ = "unknown release"
try:
    import sys
    if sys.platform == "win32":
        from gxps.xdg import DATA_DIR
        with open(DATA_DIR / "version.txt") as vfile:
            __version__ = vfile.readline().strip()
        with open(DATA_DIR / "release.txt") as rfile:
            __release__ = rfile.readline().strip()
    else:
        from pbr.version import VersionInfo
        __version__ = VersionInfo("gxps").version_string()
        __release__ = VersionInfo("gxps").release_string()
except ImportError:
    __version__ = "devel"
    __release__ = "devel"
except FileNotFoundError:
    __version__ = "missingversion"
    __release__ = "missingrelease"



"""Modified GTK3+ widgets for inclusion in the glade catalog."""
# pylint: disable=wrong-import-position
# pylint: disable=logging-format-interpolation

import logging

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, Pango
import numpy as np
from matplotlib.backends.backend_gtk3agg import FigureCanvasGTK3Agg
from matplotlib.backends.backend_gtk3 import NavigationToolbar2GTK3
from matplotlib.figure import Figure

from gxps import __appname__, __release__, __authors__, __website__
from gxps.config import COLORS, CONFIG
from gxps.state import State
from gxps.canvas_tools import (
    PeakSelector, SpanSelector, PointSelector,
    DraggableVLine
    )


LOG = logging.getLogger(__name__)



class GXPSAppWindow(Gtk.ApplicationWindow):
    """Main window of GXPS.
    """
    __gtype_name__ = "GXPSAppWindow"
    def __init__(self):
        LOG.info("Instantiating window")
        super().__init__()

    def startup(self, app):
        """Position the window correctly."""
        self.set_application(app)
        xpos, ypos = CONFIG["Window"]["xpos"], CONFIG["Window"]["ypos"]
        xsize, ysize = CONFIG["Window"]["xsize"], CONFIG["Window"]["ysize"]
        if xpos and ypos:
            self.move(int(xpos), int(ypos))
        if xsize and ysize:
            self.set_default_size(int(xsize), int(ysize))
        self.set_wmclass(__appname__, __appname__)
        super().show_all()


class GXPSExportCanvas(FigureCanvasGTK3Agg):
    """Canvas for drawing the spectra prior to export.
    """
    # pylint: disable=invalid-name
    __gtype_name__ = "GXPSExportCanvas"
    def __init__(self):
        figure = Figure()
        super().__init__(figure)
        self._ax = self.figure.add_subplot(111)
        self._ax.invert_xaxis()

    @property
    def ax(self):
        """Expose axis object."""
        return self._ax

    def _set_ticks(self):
        """Configures axes ticks.
        """
        self.ax.tick_params(
            reset=True,
            axis="both",
            direction="in",
            labelsize="large",
            labelcolor=COLORS["Image"]["axisticks"],
            color=COLORS["Image"]["axisticks"],
            labelleft=False
        )

    def plot_spectra(self, spectra):
        """Plots given spectra."""
        for spectrum in spectra:
            self._ax.plot(spectrum.energy, spectrum.intensity)
            if spectrum.background.any():
                self._ax.plot(spectrum.energy, spectrum.background)
            if spectrum.fit.any():
                intensity = spectrum.background + spectrum.fit
                self._ax.plot(spectrum.energy, intensity)
            for peak in spectrum.peaks:
                intensity = spectrum.background + peak.intensity
                self._ax.plot(spectrum.energy, intensity, ls="-")

    def saveas(self, fname, **kwargs):
        """Saves the plot to an image file."""
        self.figure.savefig(fname, **kwargs)


class GXPSCanvas(FigureCanvasGTK3Agg):
    """Canvas for drawing the spectra.
    """
    # pylint: disable=invalid-name
    __gtype_name__ = "GXPSCanvas"
    def __init__(self):
        figure = Figure()
        super().__init__(figure)
        self.ax = self.figure.add_axes([-0.005, 0.05, 1.01, 1.005])
        self.ax.set_facecolor(COLORS["Canvas"]["canvas-facecolor"])
        self.figure.patch.set_facecolor(COLORS["Canvas"]["canvas-facecolor"])
        self._xy_buffer = [0, 1, 0, 1]
        self._xy_center = [np.inf, -np.inf, np.inf, -np.inf]
        self._set_ticks()

    def store_xylims(self):
        """Stores axis limits in self._xy_buffer.
        """
        xmin, xmax = min(self.ax.get_xlim()), max(self.ax.get_xlim())
        ymin, ymax = min(self.ax.get_ylim()), max(self.ax.get_ylim())
        self._xy_buffer = [xmin, xmax, ymin, ymax]

    def restore_xylims(self):
        """Sets the axis limits to the ones stored in self._xy_buffer.
        """
        if not np.all(np.isfinite(self._xy_buffer)):
            self._xy_buffer = [0, 1, 0, 1]
        xmin, xmax, ymin, ymax = self._xy_buffer
        self.ax.set_xlim(xmax, xmin)
        self.ax.set_ylim(ymin, ymax)
        self._set_ticks()

    def center_view(self):
        """Focuses view on current plot. by setting self._xy_buffer to a
        frame around self._xy_center and then applying those values.
        """
        if np.all(np.isfinite(self._xy_center)):
            xspan = self._xy_center[1] - self._xy_center[0]
            bordered_center = [
                self._xy_center[0] - xspan * 0.02,
                self._xy_center[1] + xspan * 0.02,
                0,
                self._xy_center[3] * 1.1
            ]
            self._xy_buffer = bordered_center
        else:
            self._xy_buffer = [0, 1, 0, 1]
        self.restore_xylims()

    def reset_xy_centerlims(self):
        """Sets self._xy to infinite values again, so it can be incrementally
        updated by update_xy.
        """
        self._xy_center = [np.inf, -np.inf, np.inf, -np.inf]

    def update_xy_centerlims(self, xmin, xmax, ymin, ymax):
        """Updates self._xy where min values are only assumed when they
        are lower than the current min. Analogous for max values.
        """
        self._xy_center = [
            min(self._xy_center[0], xmin),
            max(self._xy_center[1], xmax),
            min(self._xy_center[2], ymin),
            max(self._xy_center[3], ymax)
        ]

    def _set_ticks(self):
        """Configures axes ticks.
        """
        self.ax.spines["bottom"].set_visible(False)
        self.ax.tick_params(
            reset=True,
            axis="both",
            direction="out",
            # pad=-20,
            labelsize="large",
            labelcolor=COLORS["Plotting"]["axisticks"],
            color=COLORS["Plotting"]["axisticks"],
            labelleft=False,
            top=False,
            left=False,
            right=False,
            bottom=False
        )
        if self._xy_center[0] == np.inf:
            self.ax.tick_params(
                which="both",
                bottom=False,
                top=False,
                left=False,
                right=False,
                labelbottom=False
            )


class DraggableVLineContainer(Gtk.DrawingArea):
    """Contains all active and usable draggable vlines. Provides methods
    for adding new ones and connecting them to the bus.
    This object is invisibly added in the same Gtk.Box where the canvas
    is."""
    __gtype_name__ = "GXPSCanvasObjects"

    def __init__(self):
        self.bus = None
        self._lines = []

    def set_bus(self, bus):
        """Set the bus."""
        self.bus = bus

    def add_line(self, line, spectrum):
        """Add a new line and attach the bus to it."""
        if not self.bus:
            raise RuntimeError("DraggableVLineContainer has not bus")
        dline = DraggableVLine(line, spectrum)
        self._lines.append(dline)
        dline.register_queue(self.bus)

    def clear(self):
        """Remove all lines."""
        self._lines.clear()


class GXPSEditSpectrumDialog(Gtk.Dialog):
    """Dialog for editing metadata.
    """
    __gtype_name__ = "GXPSEditSpectrumDialog"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._getters = {}
        # A key that indicates fields which are not read for new values.
        self.exclusion_key = ""

    def flush(self):
        """Removes all children except last, which is the action area."""
        for child in self.get_content_area().get_children()[:-1]:
            self.get_content_area().remove(child)

    def get_values(self):
        """Returns a dictionary containing the new values."""
        values = {}
        for attr, getter in self._getters.items():
            value = getter()
            if self.exclusion_key not in value:
                values[attr] = value
        return values

    def apply(self, *_args):
        """Sends APPLY response."""
        self.emit("response", Gtk.ResponseType.APPLY)

    def add_editor_row(self, attr, labeltext, entrytext):
        """Builds a row for editing an attribute. Returns a method for
        getting the new value.
        """
        box = Gtk.Box(margin_top=5, margin_bottom=5)
        label = Gtk.Label(
            label=labeltext,
            width_chars=25,
            max_width_chars=25,
            wrap=True,
            wrap_mode=Pango.WrapMode.WORD,
            justify=Gtk.Justification.CENTER
        )
        entry = Gtk.Entry(text=entrytext, margin_left=10, margin_right=10)
        box.pack_start(label, False, False, 0)
        box.pack_start(entry, True, True, 0)
        box.show_all()
        self.get_content_area().pack_start(box, False, False, 0)
        entry.connect("activate", self.apply)
        self._getters[attr] = entry.get_text

    def add_non_editable_row(self, labeltext, valuetext):
        """Adds a row for attributes associated with the files in editing.
        Gtk.Labels instead of Gtk.Entrys because this cant be changed.
        """
        box = Gtk.Box(margin_top=10, margin_bottom=5)
        label = Gtk.Label(
            label=labeltext,
            width_chars=25,
            max_width_chars=25,
            wrap=True,
            wrap_mode=Pango.WrapMode.WORD,
            justify=Gtk.Justification.CENTER
        )
        label2 = Gtk.Label(
            label=valuetext,
            margin_left=10,
            margin_right=10,
            max_width_chars=80,
            wrap=True,
            wrap_mode=Pango.WrapMode.CHAR,
            ellipsize=Pango.EllipsizeMode.START
        )
        box.pack_start(label, False, False, 0)
        box.pack_start(label2, True, True, 0)
        box.show_all()
        self.get_content_area().pack_start(box, False, False, 0)


class GXPSSaveConfirmationDialog(Gtk.Dialog):
    """Dialog that asks if you really want to quit without saving."""
    __gtype_name__ = "GXPSSaveConfirmationDialog"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.accels = Gtk.AccelGroup()
        self.add_accel_group(self.accels)

    def set_accels(self):
        """Sets the accelerators for the buttons."""
        accels = {
            "<Control>s": Gtk.ResponseType.YES,
            "<Control>d": Gtk.ResponseType.NO,
            "<Control>c": Gtk.ResponseType.CANCEL
        }
        for accel, response in accels.items():
            key, mod = Gtk.accelerator_parse(accel)
            button = self.get_widget_for_response(response)
            button.add_accelerator(
                "activate",
                self.accels,
                key,
                mod,
                Gtk.AccelFlags.VISIBLE
            )


class GXPSAboutDialog(Gtk.AboutDialog):
    """Pre-filled about dialog. Uses info from __init__.py variables.
    """
    __gtype_name__ = "GXPSAboutDialog"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_program_name(__appname__)
        self.set_version("Version: {}".format(__release__))
        self.set_authors(__authors__)
        self.set_website(__website__)
        self.set_license_type(Gtk.License.GPL_3_0)
        commentstring = (
            "If you encounter any bugs, mail me or open an "
            "issue on github. Please include a logfile "
            "(access via Help -> View logfile)."
        )
        self.set_comments(commentstring)


class FileFilter(Gtk.FileFilter):
    """Very simple FileFilter for FileChooserDialogs."""
    def __init__(self, name, patterns):
        super().__init__()
        for pattern in patterns:
            self.add_pattern(pattern)
        self.set_name(name)


class GXPSImportDialog(Gtk.FileChooserDialog):
    """File chooser dialog for spectrum importing."""
    __gtype_name__ = "GXPSImportDialog"
    def __init__(self, *_args, **_kwargs):
        super().__init__(
            "Import data...",
            None,
            Gtk.FileChooserAction.OPEN,
            ("_Cancel", Gtk.ResponseType.CANCEL, "_Open", Gtk.ResponseType.OK),
        )
        self.set_select_multiple(True)
        self.add_filter(FileFilter("all files", ["*.xym", "*txt", "*.xy"]))
        self.add_filter(FileFilter(".xym", ["*.xym"]))
        self.add_filter(FileFilter(".xy", ["*.xy"]))
        self.add_filter(FileFilter(".txt", ["*.txt"]))


class GXPSOpenProjectDialog(Gtk.FileChooserDialog):
    """File chooser dialog for spectrum importing."""
    __gtype_name__ = "GXPSOpenProjectDialog"
    def __init__(self, *_args, **_kwargs):
        super().__init__(
            "Open project...",
            None,
            Gtk.FileChooserAction.OPEN,
            ("_Cancel", Gtk.ResponseType.CANCEL, "_Open", Gtk.ResponseType.OK),
        )
        self.add_filter(FileFilter(".gxps", ["*.gxps"]))


class GXPSMergeProjectDialog(Gtk.FileChooserDialog):
    """File chooser dialog for spectrum importing."""
    __gtype_name__ = "GXPSMergeProjectDialog"
    def __init__(self, *_args, **_kwargs):
        super().__init__(
            "Merge project...",
            None,
            Gtk.FileChooserAction.OPEN,
            ("_Cancel", Gtk.ResponseType.CANCEL, "_Open", Gtk.ResponseType.OK),
        )
        self.add_filter(FileFilter(".gxps", ["*.gxps"]))


class GXPSSaveProjectDialog(Gtk.FileChooserDialog):
    """File chooser dialog for spectrum importing."""
    __gtype_name__ = "GXPSSaveProjectDialog"
    def __init__(self, *_args, **_kwargs):
        super().__init__(
            "Save project...",
            None,
            Gtk.FileChooserAction.SAVE,
            ("_Cancel", Gtk.ResponseType.CANCEL, "_Save", Gtk.ResponseType.OK),
        )
        self.add_filter(FileFilter(".gxps", ["*.gxps"]))
        self.set_do_overwrite_confirmation(True)



class GXPSExportTXTDialog(Gtk.FileChooserDialog):
    """File chooser dialog for spectrum importing."""
    __gtype_name__ = "GXPSExportTXTDialog"
    def __init__(self, *_args, **_kwargs):
        super().__init__(
            "Export as textfile...",
            None,
            Gtk.FileChooserAction.SAVE,
            ("_Cancel", Gtk.ResponseType.CANCEL, "_Save", Gtk.ResponseType.OK),
        )
        self.add_filter(FileFilter(".txt", ["*.txt", "*.csv"]))
        self.set_do_overwrite_confirmation(True)


class GXPSExportIMGDialog(Gtk.FileChooserDialog):
    """File chooser dialog for spectrum importing."""
    __gtype_name__ = "GXPSExportIMGDialog"
    def __init__(self, *_args, **_kwargs):
        super().__init__(
            "Export as image...",
            None,
            Gtk.FileChooserAction.SAVE,
            ("_Cancel", Gtk.ResponseType.CANCEL, "_Save", Gtk.ResponseType.OK),
        )
        self.add_filter(FileFilter(
            "raster image",
            ["*.png", "*.jpeg", "*.jpg"]
        ))
        self.add_filter(FileFilter(
            "vector image",
            ["*.pdf", "*.svg"]
        ))
        self.add_filter(FileFilter(
            "all images",
            ["*.png, *.jpeg", "*.jpg", "*.svg", "*.pdf"]
        ))
        self.add_filter(FileFilter("all", ["*"]))
        self.set_do_overwrite_confirmation(True)


class GXPSPeakTreeStore(Gtk.TreeStore):
    """Treestore with peak information.
    """
    __gtype_name__ = "GXPSPeakTreeStore"
    def __init__(self, *_args, **_kwargs):
        titles = State.titles["peak_view"]
        # the first column is "is_actve"
        types = [object] + [str] * len(titles)
        super().__init__(*types)


class GXPSSpectrumTreeStore(Gtk.TreeStore):
    """Treestore with spectrum information.
    """
    __gtype_name__ = "GXPSSpectrumTreeStore"
    def __init__(self, *_args, **_kwargs):
        titles = State.titles["spectrum_view"]
        # the first column is "is_actve"
        types = [object] + [str] * len(titles)
        super().__init__(*types)


class Cursors():
    """Simple namespace for cursor reference.
    """
    # pylint: disable=too-few-public-methods
    HAND, POINTER, SELECT_REGION, MOVE, WAIT, DRAG, DELETE = list(range(7))


class GXPSPlotToolbar(NavigationToolbar2GTK3, Gtk.Toolbar):
    """Matplotlib-like toolbar with a few nice extra tools.
    """
    # pylint: disable=invalid-name
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=super-init-not-called
    __gtype_name__ = "GXPSPlotToolbar"
    def __init__(self, *args, **kwargs):
        Gtk.Toolbar.__init__(self, *args, **kwargs)
        self.cursors = Cursors()
        self.cursord = {
            self.cursors.MOVE: Gdk.Cursor.new(Gdk.CursorType.FLEUR),
            self.cursors.HAND: Gdk.Cursor.new(Gdk.CursorType.HAND2),
            self.cursors.POINTER: Gdk.Cursor.new(Gdk.CursorType.LEFT_PTR),
            self.cursors.SELECT_REGION: Gdk.Cursor.new(Gdk.CursorType.TCROSS),
            self.cursors.WAIT: Gdk.Cursor.new(Gdk.CursorType.WATCH),
            self.cursors.DRAG: Gdk.Cursor.new(
                Gdk.CursorType.SB_H_DOUBLE_ARROW),
            self.cursors.DELETE: Gdk.Cursor.new(Gdk.CursorType.X_CURSOR)
        }
        self._centerlims = [np.inf, -np.inf, np.inf, -np.inf]

    def startup(self, builder):
        """Call the real super __init__ after the canvas is known and give
        it to this function.
        """
        # pylint: disable=attribute-defined-outside-init
        self.message = builder.get_object("mpl_message_label")
        self.mode_label = self.message
        self.xy_label = builder.get_object("mpl_coord_label")
        canvas = builder.get_object("main_canvas")
        self.toggle_buttons = {
            "Add region": builder.get_object("region_add_button"),
            "Remove region": builder.get_object("region_remove_button"),
            "Add peak": builder.get_object("peak_add_button"),
            "Pan / Zoom": builder.get_object("mpl_pan_button"),
            "Zoom Rectangle": builder.get_object("mpl_zoom_button")
        }
        super().__init__(canvas, None)

        self.span_selector = SpanSelector(
            self.canvas.figure.get_axes()[0],
            lambda *args: None,
            "horizontal",
            minspan=0.2,
            span_stays=False,
            useblit=True
        )
        self.span_selector.active = False
        self.peak_selector = PeakSelector(
            self.canvas.figure.get_axes()[0],
            lambda *args: None,
            peak_stays=False,
            useblit=True
        )
        self.peak_selector.active = False
        self.point_selector = PointSelector(
            self.canvas.figure.get_axes()[0],
            lambda *args: None,
        )
        self.point_selector.active = False

    def _init_toolbar(self):
        """Normally, this would create the buttons and connect them to
        the tools, but now GtkBuilder does the job and the connections
        are done in the Gtk.Application. This function is automatically
        called during NavigationToolbar.__init__
        """

    def set_history_buttons(self):
        """Same as above, this prevents the use of self._gtk_ids."""

    def untoggle_buttons(self, exclude=None):
        """Untoggles all buttons (exclude the one given by its label)."""
        if exclude is None:
            exclude = []
        for label, button in self.toggle_buttons.items():
            if label not in exclude:
                button.set_active(False)

    def disable_tools(self):
        """Release widgetlock and disconnect all signals associated with
        native matplotlib toolbar tools.
        """
        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)
        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)

        self._active = None
        self.mode = ""
        self.untoggle_buttons()
        self.release_all_tools()
        for ax in self.canvas.figure.get_axes():
            ax.set_navigate_mode(self._active)
        self.set_message(self.mode)

    def release_all_tools(self):
        """Release all tools used in this class from self.canvas.widgetlock
        because the locks have different owners.
        """
        try:
            self.canvas.widgetlock.release(self)
        except ValueError:
            try:
                self.canvas.widgetlock.release(self.span_selector)
            except ValueError:
                try:
                    self.canvas.widgetlock.release(self.peak_selector)
                except ValueError:
                    self.canvas.widgetlock.release(self.point_selector)

    def get_point(self, callback, onmove_callback=None):
        """Gets a x, y point on the canvas and then calls
        callback(x0, y0, x, y) where x0, y0 are the coords when pressing
        the mouse and x, y are the coords when releasing.
        """
        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)
            self.mode = ""
        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
            self.mode = ""

        mode = "Remove region"

        self.release_all_tools()
        if self._active == mode:
            self._active = None
            self.mode = ""
            self.point_selector.active = False
            self.set_message(self.mode)
            return

        self._active = mode
        self.mode = mode
        self.untoggle_buttons(exclude=[mode])
        self.canvas.widgetlock(self.point_selector)
        for ax in self.canvas.figure.get_axes():
            ax.set_navigate_mode(None)
        self.set_message(self.mode)

        def on_selected(x0, y0, x, y):
            """Callback caller."""
            self._active = None
            self.mode = ""
            self.untoggle_buttons()
            self.point_selector.active = False
            self.release_all_tools()
            self.set_message(self.mode)
            self._set_cursor(event=None)
            callback(x0, y0, x, y)

        self.point_selector.onselect = on_selected
        self.point_selector.onmove_callback = onmove_callback
        self.point_selector.active = True

    def get_span(self, callback, **kwargs):
        """Gets a span and then calls callback(min, max). Also takes care
        of widgetlock and such.
        """
        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)
            self.mode = ""
        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
            self.mode = ""

        mode = "Add region"

        self.release_all_tools()
        if self._active == mode:
            self._active = None
            self.mode = ""
            self.span_selector.active = False
            self.set_message(self.mode)
            return

        self._active = mode
        self.mode = mode
        self.untoggle_buttons(exclude=[mode])
        self.canvas.widgetlock(self.span_selector)
        for ax in self.canvas.figure.get_axes():
            ax.set_navigate_mode(None)
        self.set_message(self.mode)

        def on_selected(emin, emax):
            """Callback caller."""
            self._active = None
            self.mode = ""
            self.untoggle_buttons()
            self.span_selector.active = False
            self.release_all_tools()
            self.set_message(self.mode)
            self._set_cursor(event=None)
            callback(emin, emax)
        rectprops = {
            "alpha": 1,
            "fill": False,
            "edgecolor": "black",
            "linewidth": 1,
            "linestyle": "-"
        }
        rectprops.update(kwargs)
        self.span_selector.set_rectprops(rectprops)
        self.span_selector.onselect = on_selected
        self.span_selector.active = True

    def get_wedge(self, callback, **kwargs):
        """Gets a wegde and then calls callback(center, height, fwhm).
        Also takes care of widgetlock and such.
        """
        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)
            self.mode = ""
        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
            self.mode = ""

        mode = "Add peak"

        self.release_all_tools()
        if self._active == mode:
            self._active = None
            self.mode = ""
            self.peak_selector.active = False
            self.set_message(self.mode)
            return

        self._active = mode
        self.mode = mode
        self.untoggle_buttons(exclude=[mode])
        self.canvas.widgetlock(self.peak_selector)
        for ax in self.canvas.figure.get_axes():
            ax.set_navigate_mode(None)
        self.set_message(self.mode)

        def on_selected(center, height, angle):
            """Callback caller."""
            self._active = None
            self.mode = ""
            self.untoggle_buttons()
            self.peak_selector.active = False
            self.release_all_tools()
            self.set_message(self.mode)
            self._set_cursor(event=None)
            callback(center, height, angle)
        wedgeprops = {
            "alpha": 0.5,
            "fill": True,
            "edgecolor": "black",
            "linewidth": 1,
            "linestyle": "-"
        }
        if "limits" in kwargs:
            self.peak_selector.set_limits(kwargs.pop("limits"))
        else:
            self.peak_selector.set_limits((-np.inf, np.inf))
        wedgeprops.update(kwargs)
        self.peak_selector.set_wedgeprops(wedgeprops)
        self.peak_selector.onselect = on_selected
        self.peak_selector.active = True

    def pan(self, *_ignore):
        """Activate the pan/zoom tool. pan with left button, zoom with right
        OVERWRITE because of widgetlock release.
        """
        # set the pointer icon and button press funcs to the
        # appropriate callbacks
        if self._active == 'PAN':
            self._active = None
        else:
            self._active = 'PAN'
        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)
            self.mode = ''
        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
            self.mode = ''

        mode = "Pan / Zoom"
        self.release_all_tools()        # changed here
        self.untoggle_buttons(exclude=[mode])

        def local_release_pan(*args, **kwargs):
            """Use the super release_pan, but also untoggle the button."""
            self.release_pan(*args, **kwargs)
            self.untoggle_buttons()

        if self._active:
            self._idPress = self.canvas.mpl_connect(
                'button_press_event', self.press_pan)
            self._idRelease = self.canvas.mpl_connect(
                'button_release_event', local_release_pan)
            self.mode = mode
            self.canvas.widgetlock(self)
        for ax in self.canvas.figure.get_axes():
            ax.set_navigate_mode(self._active)
        self.set_message(self.mode)

    def zoom(self, *args):
        """Activate zoom to rect mode.
        OVERWRITE because of widgetlock release.
        """
        if self._active == 'ZOOM':
            self._active = None
        else:
            self._active = 'ZOOM'
        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)
            self.mode = ''
        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
            self.mode = ''

        mode = "Zoom Rectangle"
        self.release_all_tools()        # changed here
        self.untoggle_buttons(exclude=[mode])

        def local_release_zoom(*args, **kwargs):
            """Use the super release_zoom, but also untoggle the button."""
            self.release_zoom(*args, **kwargs)
            self.untoggle_buttons()

        if self._active:
            self._idPress = self.canvas.mpl_connect(
                'button_press_event', self.press_zoom)
            self._idRelease = self.canvas.mpl_connect(
                'button_release_event', local_release_zoom)
            self.mode = mode
            self.canvas.widgetlock(self)
        for ax in self.canvas.figure.get_axes():
            ax.set_navigate_mode(self._active)
        self.set_message(self.mode)

    def draw_rubberband(self, event, x0, y0, x1, y1):
        # pylint: disable=too-many-arguments
        """adapted from
        http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/189744
        """
        self.ctx = self.canvas.get_property("window").cairo_create()
        self.canvas.draw()
        height = self.canvas.figure.bbox.height
        y1 = height - y1
        y0 = height - y0
        w = abs(x1 - x0)
        h = abs(y1 - y0)
        rect = [int(val) for val in (min(x0, x1), min(y0, y1), w, h)]
        self.ctx.new_path()
        self.ctx.set_line_width(0.5)
        self.ctx.rectangle(rect[0], rect[1], rect[2], rect[3])
        self.ctx.set_source_rgb(0.95, 0.95, 0.95)   # this is the change
        self.ctx.stroke()

    def center(self):
        """Centers view and disables navbar tools.
        """
        self.disable_tools()
        self.untoggle_buttons()
        self.canvas.center_view()
        self.canvas.draw_idle()

    def mouse_move(self, event):
        self._set_cursor(event)
        if event.inaxes and event.inaxes.get_navigate():
            try:
                s = event.inaxes.format_coord(event.xdata, event.ydata)
                self.xy_label.set_markup("<tt>{}</tt>".format(s))
            except (ValueError, OverflowError):
                pass
        if self.mode:
            self.mode_label.set_text("{}".format(self.mode))

    def set_cursor(self, cursor):
        self.canvas.get_property("window").set_cursor(self.cursord[cursor])
        Gtk.main_iteration()

    def _set_cursor(self, event):
        if not event or not event.inaxes or not self._active:
            if self._lastCursor != self.cursors.POINTER:
                self.set_cursor(self.cursors.POINTER)
                self._lastCursor = self.cursors.POINTER
        else:
            if (self._active == 'ZOOM'
                    and self._lastCursor != self.cursors.SELECT_REGION):
                self.set_cursor(self.cursors.SELECT_REGION)
                self._lastCursor = self.cursors.SELECT_REGION
            elif (self._active == 'PAN' and
                  self._lastCursor != self.cursors.MOVE):
                self.set_cursor(self.cursors.MOVE)
                self._lastCursor = self.cursors.MOVE
            elif (self._active == 'Add region' and
                  self._lastCursor != self.cursors.DRAG):
                self.set_cursor(self.cursors.DRAG)
                self._lastCursor = self.cursors.DRAG
            elif (self._active == 'Remove region' and
                  self._lastCursor != self.cursors.DELETE):
                self.set_cursor(self.cursors.DELETE)
                self._lastCursor = self.cursors.DELETE
            elif self._active == "Add peak":
                lims = self.peak_selector.limits
                if lims[0] < event.xdata < lims[1]:
                    if self._lastCursor != self.cursors.SELECT_REGION:
                        self.set_cursor(self.cursors.SELECT_REGION)
                        self._lastCursor = self.cursors.SELECT_REGION
                else:
                    if self._lastCursor != self.cursors.POINTER:
                        self.set_cursor(self.cursors.POINTER)
                        self._lastCursor = self.cursors.POINTER



"""Classes for issuing commands and managing undo/redo history."""
# pylint: disable=logging-format-interpolation

import logging

from gxps.controller import File, Edit, Fit, ViewC, Help


LOG = logging.getLogger(__name__)


class CommandSender:
    """Gets all user input and triggers the appropriate functions."""
    # pylint: disable=too-many-instance-attributes
    def __init__(self, bus, data, state, get_widget):
        self.bus = bus
        self.bus.set_policy("accumulate")
        self.data = data
        self.state = state
        self.get_widget = get_widget
        self.history = History()

        self._file = File(get_widget, self.state, self.data, self.bus)
        self._edit = Edit(get_widget, self.state, self.data, self.bus)
        self._fit = Fit(get_widget, self.state, self.data, self.bus)
        self._view = ViewC(get_widget, self.state, self.data, self.bus)
        self._help = Help(get_widget, self.state, self.data, self.bus)

        self.callbacks = {
            # File actions
            "startup": self._file.startup,
            "ask-for-save": self._file.ask_for_save,
            "project-new": self._file.on_new,
            "save-project": self._file.on_save,
            "save-project-as": self._file.on_save_as,
            "open-project": self._file.on_open,
            "merge-project": self._file.on_merge,
            "import-spectra": self._file.on_import,
            "export-txt": self._file.on_export_txt,
            "export-image": self._file.on_start_image_exporter,
            "on_img_export_change": self._file.on_change_image_exporter,

            # Edit actions
            "edit-spectra": self._edit.on_edit_spectra,
            "remove-spectra": self._edit.on_remove_selected_spectra,
            "avg-selected-spectra": lambda *x: None,
            # Edit handlers
            "on_calibration_spinbutton_value_changed": self._edit.on_calibrate,
            "on_normalization_combo_changed": self._edit.on_normalize,
            "on_normalization_entry_activate": self._edit.on_normalize_manual,
            "on_region_background_type_combo_changed": self._edit.on_change_bg,

            # Fit actions
            "add-region": self._fit.on_add_region,
            "remove-region": self._fit.on_remove_region,
            "clear-regions": lambda *x: None,
            "add-peak": self._fit.on_add_peak,
            "add-guessed-peak": lambda *x: None,
            "remove-peak": self._fit.on_remove_active_peak,
            "clear-peaks": self._fit.on_clear_peaks,
            "fit": self._fit.on_fit,
            # Fit handlers
            "on_peak_entry_activate": self._fit.on_peak_entry_activate,
            "on_peak_name_entry_changed": self._fit.on_peak_name_entry_changed,
            "on_peak_model_combo_changed": self._fit.on_peak_model_changed,

            # View actions
            "show-selected-spectra": self._view.on_show_selected_spectra,
            "show-atomlib": self._view.on_show_rsfs,
            "center-plot": self._view.on_center_plot,
            "pan-plot": self._view.on_pan_plot,
            "zoom-plot": self._view.on_zoom_plot,
            # View handlers
            "on_spectrum_view_search_entry_changed":
                self._view.on_spectrum_view_filter_changed,
            "on_spectrum_view_search_combo_changed":
                self._view.on_spectrum_view_filter_changed,
            "on_spectrum_view_button_press_event":
                self._view.on_spectrum_view_clicked,
            "on_spectrum_view_row_activated":
                self._view.on_spectrum_view_row_activated,
            "on_peak_view_row_activated":
                self._view.on_peak_view_row_activated,

            # Help actions
            "view-logfile": self._help.on_view_logfile,
            "edit-colors": self._help.on_edit_colors,
            "about": self._help.on_about,
        }

        self.bus.subscribe(self._fit.on_change_region, "changed-vline", 5)

    def __call__(self, *args):
        key = args[-1]
        if key not in self.callbacks:
            LOG.warning("Action/Handler {} does not exist".format(key))
            return False
        do_func = self.callbacks[key]
        command = Command(do_func)
        rval = self.history.do_command(command, *args[:-1])
        return rval

    def callback(self, *_args):
        """Manages observer-style subscriptions."""

    def execute(self, command):
        """Execute a command object."""
        self.history.do_command(command)
        self.bus.fire()


class History:
    """Stores commands, organizes undo/redo."""
    def __init__(self):
        self._commands_done = []
        self._commands_redoable = []

    def do_command(self, command, *args, **kwargs):
        """Executes a command."""
        if command.undoable:
            self._commands_redoable.clear()
        return_value = command(*args, **kwargs)
        if command.undoable:
            self._commands_done.append(command)
        return return_value

    def undo(self):
        """Undoes last command."""
        command = self._commands_done.pop()
        command.undo()
        self._commands_redoable.append(command)

    def redo(self):
        """Redoes last command."""
        command = self._commands_redoable.pop()
        command.redo()
        self._commands_done.append(command)


class Command:
    """Executes user action: Affects the self.data and self.state objects,
    uses self.get_widget.
    """
    def __init__(self, func=None, undo_func=None, redo_func=None, data=None):
        self.func = func
        if self.func is None:
            raise ValueError("Command missing executor function")
        self.undo_func = undo_func
        self.redo_func = redo_func
        self.data = data
        self.state = []
        self.data = []
        self.obj = None

    @property
    def undoable(self):
        """Make testing for undoability available."""
        return self.undo_func is not None

    @property
    def redoable(self):
        """Make testing for redoability available."""
        return self.redo_func is not None

    def __call__(self, *args, **kwargs):
        self.state = self.func(*args, **kwargs)
        return self.state

    def undo(self):
        """Call the undo function."""
        if self.undo_func is None:
            raise AttributeError("Command {} has no undoer".format(self))
        self.undo_func(*self.state)

    def redo(self, *args, **kwargs):
        """Call the redo function. If it is None, call the normal executor."""
        if self.undo_func is None:
            self(*args, **kwargs)
        self.redo_func(*self.state)



"""Tests the modules __main__ and __init__ to verify correct application
startup."""
# pylint: disable=invalid-name
# pylint: disable=missing-docstring

import pytest

from gxps.main import main

def test_main():
    with pytest.raises(SystemExit):
        main()
